/*
*
*	Example implementation for "Analytic Spectral Integration of Birefringence-Induced Iridescence"
*		by Shlomi Steinberg
*
* This code is licensed under the MIT license (MIT) (http://opensource.org/licenses/MIT)
*/


#version 450

#extension GL_ARB_separate_shader_objects : enable
#extension GL_ARB_shading_language_420pack : enable

struct Material {
	vec4 albedo;
	vec4 A;
	vec4 mu_o;
	vec4 mu_e;
	float eta_o;
	float eta_e;
	float _unused0, _unused1;
};

layout (binding = 0) uniform restrict UBO {
	mat4 projection;
	mat4 modelView;
	mat4 invModelView;
	vec4 cameraPosition;
	vec2 framebuffer_size;
	int envprobe;
	float polarization;
	float noise_amount;
} ubo;
layout (binding = 1) uniform samplerCube samplerEnv[3];
layout (binding = 3) uniform sampler2D spectralLut;

layout (constant_id = 0) const int view_mode = 0;
layout (constant_id = 1) const int m = 1;
layout (constant_id = 2) const int inject_noise = 0;

layout (binding = 2) uniform restrict Params {
	Material material;
	float exposure;
	float d;
} params;

layout (location = 0) in vec3 color;
layout (location = 1) in vec3 position;
layout (location = 2) in vec2 uv;
layout (location = 3) in vec3 normal;
layout (location = 4) in vec3 wposition;

layout (location = 0) out vec4 outFragColor;

const float pi = 3.1415f;
const float eta_outside = 1.f;


// Utility
float sqr(float x) { return x*x; }
vec2 sqr(vec2 x) { return x*x; }
vec3 sqr(vec3 x) { return x*x; }
vec4 sqr(vec4 x) { return x*x; }
float max_element(vec2 v) { return max(v.x, v.y); }
float min_element(vec2 v) { return min(v.x, v.y); }
float max_element(vec3 v) { return max(v.x, max(v.y, v.z)); }
float min_element(vec3 v) { return min(v.x, min(v.y, v.z)); }
float max_element(vec4 v) { return max(max(v.x, v.y), max(v.w, v.z)); }
float min_element(vec4 v) { return min(min(v.x, v.y), min(v.w, v.z)); }
// Complex arithmetics
struct complex_mat2 { mat2 r, i; };
struct complex_vec2 { vec2 r, i; };
vec2  product(vec2 a, vec2 b) { return vec2(a.x*b.x-a.y*b.y, a.x*b.y+a.y*b.x); }
vec2  divide(vec2 a, vec2 b) { return vec2((a.x*b.x+a.y*b.y)/(b.x*b.x+b.y*b.y),(a.y*b.x-a.x*b.y)/(b.x*b.x+b.y*b.y)); }
complex_vec2 cvec2(vec2 a, vec2 b) {
	complex_vec2 v;
	v.r = vec2(a.x, b.x);
	v.i = vec2(a.y, b.y);
	return v;
}
complex_mat2 cidentity() {
	complex_mat2 R;
	R.r = mat2(1);
	R.i = mat2(0);
	return R;
}
complex_mat2 cmat2(vec2 a, vec2 b, vec2 c, vec2 d) {
	complex_mat2 R;
	R.r[0][0] = a.x;	R.i[0][0] = a.y;
	R.r[1][0] = b.x;	R.i[1][0] = b.y;
	R.r[0][1] = c.x;	R.i[0][1] = c.y;
	R.r[1][1] = d.x;	R.i[1][1] = d.y;
	return R;
}
complex_mat2 cmat2_add(complex_mat2 A, complex_mat2 B) {
	complex_mat2 R;
	R.r = A.r + B.r;
	R.i = A.i + B.i;
	return R;
}
complex_mat2 cmat2_substract(complex_mat2 A, complex_mat2 B) {
	complex_mat2 R;
	R.r = A.r - B.r;
	R.i = A.i - B.i;
	return R;
}
complex_mat2 cmat2_product(complex_mat2 A, complex_mat2 B) {
	const vec2 a = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d = vec2(A.r[1][1], A.i[1][1]);
	const vec2 a2 = vec2(B.r[0][0], B.i[0][0]);
	const vec2 b2 = vec2(B.r[1][0], B.i[1][0]);
	const vec2 c2 = vec2(B.r[0][1], B.i[0][1]);
	const vec2 d2 = vec2(B.r[1][1], B.i[1][1]);

	const vec2 Ra = product(a,a2) + product(b,c2);
	const vec2 Rb = product(a,b2) + product(b,d2);
	const vec2 Rc = product(c,a2) + product(d,c2);
	const vec2 Rd = product(c,b2) + product(d,d2);

	return cmat2(Ra, Rb, Rc, Rd);
}
complex_vec2 vec2_by_cmat2(vec2 v, complex_mat2 B) {
	const vec2 a  = vec2(v.x,0);
	const vec2 b  = vec2(v.y,0);
	const vec2 a2 = vec2(B.r[0][0], B.i[0][0]);
	const vec2 b2 = vec2(B.r[1][0], B.i[1][0]);
	const vec2 c2 = vec2(B.r[0][1], B.i[0][1]);
	const vec2 d2 = vec2(B.r[1][1], B.i[1][1]);

	const vec2 Ra = product(a,a2) + product(b,c2);
	const vec2 Rb = product(a,b2) + product(b,d2);

	return cvec2(Ra, Rb);
}
complex_vec2 cvec2_by_mat2(complex_vec2 v, mat2 B) {
	const vec2 a  = vec2(v.r[0],v.i[0]);
	const vec2 b  = vec2(v.r[1],v.i[1]);
	const vec2 a2 = vec2(B[0][0], 0);
	const vec2 b2 = vec2(B[0][1], 0);
	const vec2 c2 = vec2(B[1][0], 0);
	const vec2 d2 = vec2(B[1][1], 0);

	const vec2 Ra = product(a,a2) + product(b,c2);
	const vec2 Rb = product(a,b2) + product(b,d2);

	return cvec2(Ra, Rb);
}
complex_mat2 cinv(complex_mat2 A) {
	const vec2 a = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d = vec2(A.r[1][1], A.i[1][1]);

	const vec2 N = divide(vec2(1,0), product(a,d) - product(b,c));
	const vec2 Ra = divide(d,N);
	const vec2 Rb = divide(-b,N);
	const vec2 Rc = divide(-c,N);
	const vec2 Rd = divide(a,N);

	return cmat2(Ra, Rb, Rc, Rd);
}
float mag2(vec2 v) { return sqr(v.x) + sqr(v.y); }
vec2  mag2(complex_vec2 v) { return vec2(sqr(v.r[0]) + sqr(v.i[0]), sqr(v.r[1]) + sqr(v.i[1])); }
mat2  mag2(complex_mat2 A) {
	const vec2 a = vec2(A.r[0][0], A.i[0][0]);
	const vec2 b = vec2(A.r[1][0], A.i[1][0]);
	const vec2 c = vec2(A.r[0][1], A.i[0][1]);
	const vec2 d = vec2(A.r[1][1], A.i[1][1]);

	return mat2(mag2(a), mag2(c), mag2(b), mag2(d));
}


mat3 tangent_frame(vec3 N, vec3 p, vec2 uv)
{
    // get edge vec­tors of the pix­el tri­an­gle
    vec3 dp1 = dFdx(p);
    vec3 dp2 = dFdy(p);
    vec2 duv1 = dFdx(uv);
    vec2 duv2 = dFdy(uv);

    // solve the lin­ear sys­tem
    vec3 dp2perp = cross(dp2, N);
    vec3 dp1perp = cross(N, dp1);
    vec3 T = dp2perp * duv1.x + dp1perp * duv2.x;
    vec3 B = dp2perp * duv1.y + dp1perp * duv2.y;

    // con­struct a scale-invari­ant frame 
    float invmax = inversesqrt(max(dot(T,T), dot(B,B)));
    return mat3(T * invmax, B * invmax, N);
}
vec4 mod289(vec4 x) { return x - floor(x * (1.0 / 289.0)) * 289.0; }
vec4 permute(vec4 x) { return mod289(((x*34.0)+1.0)*x); }
vec4 taylorInvSqrt(vec4 r) { return 1.79284291400159 - 0.85373472095314 * r; }
vec2 fade(vec2 t) { return t*t*t*(t*(t*6.0-15.0)+10.0); } 
// Classic Perlin noise
float cnoise(vec2 P) {
    vec4 Pi = floor(P.xyxy) + vec4(0.0, 0.0, 1.0, 1.0);
    vec4 Pf = fract(P.xyxy) - vec4(0.0, 0.0, 1.0, 1.0);
    Pi = mod289(Pi); // To avoid truncation effects in permutation
    vec4 ix = Pi.xzxz;
    vec4 iy = Pi.yyww;
    vec4 fx = Pf.xzxz;
    vec4 fy = Pf.yyww;
     
    vec4 i = permute(permute(ix) + iy);
     
    vec4 gx = fract(i * (1.0 / 41.0)) * 2.0 - 1.0 ;
    vec4 gy = abs(gx) - 0.5 ;
    vec4 tx = floor(gx + 0.5);
    gx = gx - tx;
     
    vec2 g00 = vec2(gx.x,gy.x);
    vec2 g10 = vec2(gx.y,gy.y);
    vec2 g01 = vec2(gx.z,gy.z);
    vec2 g11 = vec2(gx.w,gy.w);
     
    vec4 norm = taylorInvSqrt(vec4(dot(g00, g00), dot(g01, g01), dot(g10, g10), dot(g11, g11)));
    g00 *= norm.x;  
    g01 *= norm.y;  
    g10 *= norm.z;  
    g11 *= norm.w;  
     
    float n00 = dot(g00, vec2(fx.x, fy.x));
    float n10 = dot(g10, vec2(fx.y, fy.y));
    float n01 = dot(g01, vec2(fx.z, fy.z));
    float n11 = dot(g11, vec2(fx.w, fy.w));
     
    vec2 fade_xy = fade(Pf.xy);
    vec2 n_x = mix(vec2(n00, n01), vec2(n10, n11), fade_xy.x);
    float n_xy = mix(n_x.x, n_x.y, fade_xy.y);
    return 2.3 * n_xy;
}
float fbm(vec2 P, int octaves, float lacunarity, float gain) {
    float sum = 0.0;
    float amp = 1.0;
    vec2 pp = P;
     
    int i;
     
    for(i = 0; i < octaves; i+=1) {
        amp *= gain; 
        sum += amp * cnoise(pp);
        pp *= lacunarity;
    }
    return sum;
}
 
// Adobe XYZ to RGB
// (used as an example)
vec3 XYZtoRGB(vec3 XYZ) {
	return XYZ * mat3( 2.0413690, -0.5649464, -0.3446944,
					  -0.9692660,  1.8760108,  0.0415560,
					   0.0134474, -0.1183897,  1.0154096);
}

// Inputs:  Wavelength in nanometers
// From "Simple Analytic Approximations to the CIE XYZ Color Matching Functions"
float X_1931(float wave) {
	const float t1 = (wave-442.0f) * ((wave<442.0f)?0.0624f:0.0374f);
	const float t2 = (wave-599.8f) * ((wave<599.8f)?0.0264f:0.0323f);
	const float t3 = (wave-501.1f) * ((wave<501.1f)?0.0490f:0.0382f);
	return 0.362f * exp(-0.5f * sqr(t1)) + 1.056f * exp(-0.5f * sqr(t2)) - 0.065f * exp(-0.5f * sqr(t3));
}
float Y_1931(float wave) {
	const float t1 = (wave-568.8f) * ((wave<568.8f)?0.0213f:0.0247f);
	const float t2 = (wave-530.9f) * ((wave<530.9f)?0.0613f:0.0322f);
	return 0.821f * exp(-0.5f * sqr(t1)) + 0.286f * exp(-0.5f * sqr(t2));
}
float Z_1931(float wave) {
	const float t1 = (wave-437.0f) * ((wave<437.0f)?0.0845f:0.0278f);
	const float t2 = (wave-459.0f) * ((wave<459.0f)?0.0385f:0.0725f);
	return 1.217f * exp(-0.5f * sqr(t1)) + 0.681f * exp(-0.5f * sqr(t2));
}

// Normal modes
float qo(float eo, float K) {
    const float k = 1;
	return sqrt(sqr(eo)*sqr(k) - sqr(K));
}
vec2 qe(float eo, float ee, vec3 A, float K) {
    const float e = sqr(ee)-sqr(eo);
    const float k = 1;
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float G = sqr(b)*e+sqr(eo);
    
    const float qep = (eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(k)*sqr(ee)*G) + c*b*K*e)/G;
    const float qem = (-eo*sqrt(sqr(K)*(sqr(a)*e-sqr(ee)) + sqr(k)*sqr(ee)*G) + c*b*K*e)/G;

	return vec2(qep, qem);
}
// Electric fields
vec3 E_ep(float eo, float ee, vec3 A, float K) {
    const float k = 1;
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float q = qe(eo,ee,A,K).x;
    
    const vec3 E_e = vec3(a*sqr(k)*sqr(eo), 
						  K*c*q + b*(sqr(eo)*sqr(k) - sqr(q)), 
						  K*b*q + c*(sqr(eo)*sqr(k) - sqr(K)));
	return normalize(E_e);
}
vec3 E_em(float eo, float ee, vec3 A, float K) {
    const float k = 1;
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float q = qe(eo,ee,A,K).y;
    
    const vec3 E_e = vec3(a*sqr(k)*sqr(eo), 
						  K*c*q + b*(sqr(eo)*sqr(k) - sqr(q)), 
						  K*b*q + c*(sqr(eo)*sqr(k) - sqr(K)));
	return normalize(E_e);
}
vec3 E_op(float eo, vec3 A, float K) {
    const float k = 1;
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float q = qo(eo,K);
    
    const vec3 E_o = vec3(-b*K-c*q, a*K, a*q);
	return normalize(E_o);
}
vec3 E_om(float eo, vec3 A, float K) {
    const float k = 1;
    const float a = A.x;
    const float b = A.y;
    const float c = A.z;
    const float q = -qo(eo,K);
    
    const vec3 E_o = vec3(-b*K-c*q, a*K, a*q);
	return normalize(E_o);
}
// Poynting vector
vec3 Poynting(float q, vec3 E, float K) {
	return normalize(
		vec3(E.x*(K*E.z - q*E.y),
			 K*E.z*E.y + q*(sqr(E.x) + sqr(E.z)),
			 -q*E.z*E.y - K*(sqr(E.x) + sqr(E.y))));
}
// Reflected and refracted vectors
vec3 refracto(vec3 I, vec3 A) {
	const float K = eta_outside * I.z;
	const float eo = params.material.eta_o;

	const float q = -qo(eo,K);
	const vec3 E = E_om(eo,A,K);
	return Poynting(q,E,K);
}
vec3 refracte(vec3 I, vec3 A) {
	const float K = eta_outside * I.z;
	const float eo = params.material.eta_o;
	const float ee = params.material.eta_e;
	
	const float q = qe(eo,ee,A,K).y;
	const vec3 E = E_em(eo,ee,A,K);
	return Poynting(q,E,K);
}

// Fresnel coefficients
void fresnel_iso_aniso(float cos_theta, float sin_theta, vec3 A,
					   out float rss, out float rsp, out float tso, out float tse, out float rps, out float rpp, out float tpo, out float tpe) {
	const float ei = eta_outside;
	const float eo = params.material.eta_o;
	const float ee = params.material.eta_e;

	const float k = 1;
	const float K = k*ei*sin_theta;
	const float qi = -qo(ei,K);
	const vec3 E2om = E_om(eo,A,K);
	const vec3 E2em = E_em(eo,ee,A,K);
	const float q2om = -qo(eo,K);
	const float q2em = qe(eo,ee,A,K).y;

	const float A_1 = -E2em.x*E2om.y*K*cos_theta;
	const float A_2 = -E2em.y*E2om.x*K*cos_theta;
	const float C_1 = qi*cos_theta - K*sin_theta;
	const float C_2 = E2em.x*E2om.z*(C_1 + q2om*cos_theta) - A_1;
	const float C_3 = E2em.z*E2om.x*(C_1 + q2em*cos_theta) - A_2;

	const float Ns = C_2*(qi + q2em) - C_3*(qi + q2om);

	rps = (2*E2em.x*E2om.x*cos_theta*C_1*(q2om - q2em)) / Ns;
	rpp = (-E2em.x*E2om.z*(C_1 - q2om*cos_theta)*(qi + q2em) + E2em.z*E2om.x*(C_1 - q2em*cos_theta)*(qi + q2om) - A_1*(q2em + qi) + A_2*(q2om + qi)) / Ns;
	tpo = -(2*E2em.x*cos_theta*(q2em + qi)*C_1) / Ns;
	tpe = (2*E2om.x*cos_theta*(q2om + qi)*C_1) / Ns;

	rss = (C_2*(qi - q2em) + C_3*(q2om - qi)) / Ns;
	rsp = (2*qi*(E2em.z*E2om.y*K - E2em.y*E2om.z*K + E2em.z*E2om.z*(q2om - q2em))) / Ns;
	tso = -(2*qi*(E2em.y*K*cos_theta + E2em.z*C_1 + E2em.z*q2em*cos_theta)) / Ns;
	tse = (2*qi*(E2om.y*K*cos_theta + E2om.z*C_1 + E2om.z*q2om*cos_theta)) / Ns;

}
void fresnel_aniso_iso(float cos_phi, float sin_phi, vec3 A,
					   out float roo, out float roe, out float tos, out float top, out float reo, out float ree, out float tes, out float tep) {
	const float eo2 = eta_outside;
	const float eo = params.material.eta_o;
	const float ee = params.material.eta_e;

	const float k = 1;
	const float K = k * eta_outside * sin_phi;
	const float q2om = -qo(eo2,K);
	const float qop  =  qo(eo,K);
	const vec2 _qe = qe(eo,ee,A,K);
	const float qep = _qe.x;
	const float qem = _qe.y;
	const vec3 Eop = E_op(eo,A,K);
	const vec3 Eep = E_ep(eo,ee,A,K);
	const vec3 Eom = E_om(eo,A,K);
	const vec3 Eem = E_em(eo,ee,A,K);

	const float B_1 = -Eep.x*K*cos_phi*(q2om - qep);
	const float B_2 = -Eom.x*K*cos_phi*(q2om + qop);
	const float B_3 = -Eop.x*K*cos_phi*(qop - q2om);
	const float D_1 = K*sin_phi - q2om*cos_phi;
	const float D_2 = Eep.z*(D_1 + qep*cos_phi);
	const float D_3 = Eem.z*(D_1 + qem*cos_phi);
	const float D_4 = Eop.z*(D_1 + qop*cos_phi);

	const float N2 = (Eep.x*(qep - q2om)*D_4 + Eop.x*(q2om - qop)*D_2 + Eop.y*B_1 + Eep.y*B_3);

	roo = -(Eep.x*Eom.z*(q2om - qep)*(qop*cos_phi - D_1) + Eom.x*(q2om + qop)*D_2 + Eom.y*B_1 - Eep.y*B_2) / N2;
	roe = -(-Eom.x*Eop.z*(qop*(qop*cos_phi + K*sin_phi) + q2om*D_1) + Eom.z*Eop.x*(qop*(qop*cos_phi - K*sin_phi) + q2om*D_1) + Eop.y*B_2 + Eom.y*B_3) / N2;
	tos = (Eep.x*qep*roe - Eom.x*qop + Eop.x*qop*roo) / q2om;
	top = -(Eom.z*qop - Eom.y*K - (Eep.y*K + Eep.z*qep)*roe - (Eop.y*K + Eop.z*qop)*roo) / D_1;

	reo = (Eem.x*(qem - q2om)*(D_2 + Eep.y*K*cos_phi) + Eep.x*(q2om - qep)*(D_3 + Eem.y*K*cos_phi)) / N2;
	ree = -(Eem.x*(qem - q2om)*D_4 - Eem.x*Eop.y*K*cos_phi*(q2om - qem) + Eem.y*B_3 + Eop.x*(q2om - qop)*D_3) / N2;
	tes = (Eem.x*qem + Eep.x*qep*ree + Eop.x*qop*reo) / q2om;
	tep = -(Eem.z + Eep.z*ree + Eop.z*reo) / cos_phi;
}

// Z-displacement
float Pz(int no, int ne, int no2, int ne2, float Ioz, float Iez, float Ioz2, float Iez2) { return no*Ioz  + ne*Iez + no2*Ioz2 + ne2*Iez2; }
// Optical path length
float OPL(int no, int ne, int no2, int ne2, float OPLo, float OPLe, float OPLo2, float OPLe2) { return no*OPLo + ne*OPLe + no2*OPLo2 + ne2*OPLe2; }
// Optical path difference
float OPD(float sin_theta, int no, int ne, int no2, int ne2, float OPLo, float OPLe, float OPLo2, float OPLe2, float Ioz, float Iez, float Ioz2, float Iez2) {
	const float ei = eta_outside;
	return Pz(no, ne, no2, ne2, Ioz, Iez, Ioz2, Iez2)*sin_theta*ei + OPL(no, ne, no2, ne2, OPLo, OPLe, OPLo2, OPLe2);
}

// Analytic spectral integration
vec3 SI(float sin_theta, int no, int ne, int no2, int ne2, float OPLo, float OPLe, float OPLo2, float OPLe2, float Ioz, float Iez, float Ioz2, float Iez2) {
	const float lut_max = 0.78*3;
	const float D = abs(OPD(sin_theta, no, ne, no2, ne2, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2));
	return XYZtoRGB(texture(spectralLut, vec2(D/lut_max,0)).rgb);
}

// Numeric spectral integration, returns e^(i*k*D)
vec2 numericSI(float lambda, float sin_theta, int no, int ne, int no2, int ne2, float OPLo, float OPLe, float OPLo2, float OPLe2, float Ioz, float Iez, float Ioz2, float Iez2) {
	const float D = OPD(sin_theta, no, ne, no2, ne2, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
	const float k = 2*pi/lambda;
	return vec2(cos(k*D), sin(k*D));
}

// Comptutes the four vectors in an anistrotpic slab
void vectors_in_slab(vec3 I, vec3 A, out vec3 Io, out vec3 Io2, out vec3 Ie, out vec3 Ie2, out float e_eff, out float e_eff2) {
	const float K = eta_outside * I.z;
	const float eo = params.material.eta_o;
	const float ee = params.material.eta_e;
	const float qop = qo(eo,K);
	const float qom = -qo(eo,K);
	const float qep = qe(eo,ee,A,K).x;
	const float qem = qe(eo,ee,A,K).y;
	const vec3 Eop = E_op(eo,A,K);
	const vec3 Eom = E_om(eo,A,K);
	const vec3 Eep = E_ep(eo,ee,A,K);
	const vec3 Eem = E_em(eo,ee,A,K);
	
	Io = Poynting(qom,Eom,K);
	// Io2 = Poynting(qop,Eop,K);
	Io2 = Io * vec3(1,-1,1);
	Ie = Poynting(qem,Eem,K);
	Ie2 = Poynting(qep,Eep,K);

	// Effective refractive indices
    const vec3 We =  normalize(vec3(0,qem,-K));
    const vec3 We2 = normalize(vec3(0,qep,-K));
    e_eff = K / sqrt(1.f - sqr(We.y));
    e_eff2 = K / sqrt(1.f - sqr(We2.y));
}

void main() {
	// Slab thickness
	const float d = params.d;

	// Normal, incident and exitant vectors
	const vec3 n = normalize(cross(dFdx(wposition),dFdy(wposition)));
	const vec3 o = normalize(-position * mat3(ubo.invModelView));
	const vec3 i = normalize(reflect(-o, n));

	const float box_size = 5;
	const vec3 cube_size = 0.075f*box_size*vec3(1,1,d);		// 0.04 scale of a box with edge size of 10.
	const vec3 env_size = 0.5f*box_size*vec3(1);

	// Tranform to local space
	const vec3 Y = n;
	const vec3 X = normalize(cross(o, Y));
	const vec3 Z = cross(Y,X);
	const mat3 B = mat3(X,Y,Z);

	// Incidence vector at local frame
	const vec3 I = o * B;
	// Ordinary and extraordinary indices of refraction
	const float ei = eta_outside;
	const float eo = params.material.eta_o;
	const float ee = params.material.eta_e;
	const float K = ei * I.z;

	vec3 A = params.material.A.xzy * B;
	if (inject_noise>0) {
		// Use perlin noise to modulate the optic axis
		const vec2 modulator = 50*pow(abs(wposition.xy),vec2(7.f)) + 15*pow(abs(wposition.xy),vec2(4.7f));
		const float high_freq = fbm(pow(sin(uv),vec2(1.7f)), 4, 10.f, 0.1+.5f*params.material.A.x * params.material.A.y);
		const float noise0 = fbm(uv, 2, 4.f, params.material.A.x*(0.1+10*modulator.x)*1);
		const float noise1 = fbm(uv, 2, 3.f, params.material.A.y*(0.1+10*modulator.y)*1);
		A = normalize(vec3(noise0*0.2f + 0.6f*sin(noise0), noise1*0.2f + params.material.A.y*sin(noise1), .15)) * B;
	}


	//// Spectral integration

	// Downwards and upwards propagating ordinary and extraordinary vectors in the slab
	vec3 Io, Io2, Ie, Ie2;
	// Effeective refractive indices of the extraordinary vectors
	float e_eff, e_eff2;
	vectors_in_slab(I, A, Io, Io2, Ie, Ie2, e_eff, e_eff2);
	// OPL and z-displacement of those vectors
	const float thickness_micrometer = max(100, d * 1e+3f);
	const float sin_theta = I.z;
	const float OPLo =  abs(thickness_micrometer / Io.y) * eo;
	const float OPLe =  abs(thickness_micrometer / Ie.y) * e_eff;
	const float OPLo2 = abs(thickness_micrometer / Io2.y) * eo;
	const float OPLe2 = abs(thickness_micrometer / Ie2.y) * e_eff2;
	const float Ioz =   abs(thickness_micrometer / Io.y) * Io.z;
	const float Iez =   abs(thickness_micrometer / Ie.y) * Ie.z;
	const float Ioz2 =  abs(thickness_micrometer / Io2.y) * Io2.z;
	const float Iez2 =  abs(thickness_micrometer / Ie2.y) * Ie2.z;

	// Fresnel coefficients
	float rss, rsp, tso, tse, rps, rpp, tpo, tpe;
	float r2oo, r2oe, t2os, t2op, r2eo, r2ee, t2es, t2ep;
	float roo, roe, tos, top, reo, ree, tes, tep;
	fresnel_iso_aniso(I.y, I.z, A, rss, rsp, tso, tse, rps, rpp, tpo, tpe);
	fresnel_aniso_iso(I.y, I.z, A, r2oo, r2oe, t2os, t2op, r2eo, r2ee, t2es, t2ep);
	fresnel_aniso_iso(I.y, I.z, A*vec3(1,-1,1), roo, roe, tos, top, reo, ree, tes, tep);

	const float t1o = sqrt(mix(sqr(tso), sqr(tpo), ubo.polarization));
	const float t1e = sqrt(mix(sqr(tse), sqr(tpe), ubo.polarization));

	vec3 energy_s, energy_p;
	
	if (m == 0) {
		// Analytic

		const mat2 Dbar = transpose(mat2(sqr(r2oo)*sqr(roo) + sqr(r2oe)*sqr(reo), sqr(r2oo)*sqr(roe) + sqr(r2oe)*sqr(ree),
										 sqr(r2eo)*sqr(roo) + sqr(r2ee)*sqr(reo), sqr(r2eo)*sqr(roe) + sqr(r2ee)*sqr(ree)));
		
		// Energy
		const mat2 D = inverse(mat2(1) - Dbar);
		const vec2 Eoe = sqr(vec2(t1o, t1e)) * D;
		energy_s = vec3(dot(Eoe, sqr(vec2(t2os, t2es))));
		energy_p = vec3(dot(Eoe, sqr(vec2(t2op, t2ep))));
		// Adjust white-point
		energy_s = XYZtoRGB(energy_s);
		energy_p = XYZtoRGB(energy_p);

		// Phase shift
		// H0
		const vec3 S0 = SI(sin_theta, 1, -1, 0, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
		energy_s += 2 * S0 * abs(t1o*t2os * t1e*t2es);
		energy_p += 2 * S0 * abs(t1o*t2op * t1e*t2ep);

		// H1
		const mat2 Ts = mat2(abs(t2os), 0, 0, abs(t2es));
		const mat2 Tp = mat2(abs(t2op), 0, 0, abs(t2ep));

		const vec3 DPoo = SI(sin_theta, 1, 0, 1, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
		const vec3 DPoe = SI(sin_theta, 0, 1, 1, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
		const vec3 DPeo = SI(sin_theta, 1, 0, 0, 1, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
		const vec3 DPee = SI(sin_theta, 0, 1, 0, 1, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
		const mat2 H1r = transpose(mat2(r2oo*roo*DPoo.r + r2oe*reo*DPeo.r, r2oo*roe*DPoe.r + r2oe*ree*DPee.r,
										r2eo*roo*DPoo.r + r2ee*reo*DPeo.r, r2eo*roe*DPoe.r + r2ee*ree*DPee.r));
		const mat2 H1g = transpose(mat2(r2oo*roo*DPoo.g + r2oe*reo*DPeo.g, r2oo*roe*DPoe.g + r2oe*ree*DPee.g,
										r2eo*roo*DPoo.g + r2ee*reo*DPeo.g, r2eo*roe*DPoe.g + r2ee*ree*DPee.g));
		const mat2 H1b = transpose(mat2(r2oo*roo*DPoo.b + r2oe*reo*DPeo.b, r2oo*roe*DPoe.b + r2oe*ree*DPee.b,
										r2eo*roo*DPoo.b + r2ee*reo*DPeo.b, r2eo*roe*DPoe.b + r2ee*ree*DPee.b));

		const vec2 Eoe2 = sqr(vec2(t1o, t1e)) * D;
		energy_s.r += 2 * dot(Eoe2*Ts * H1r, abs(vec2(t2os, t2es)));
		energy_s.g += 2 * dot(Eoe2*Ts * H1g, abs(vec2(t2os, t2es)));
		energy_s.b += 2 * dot(Eoe2*Ts * H1b, abs(vec2(t2os, t2es)));
		energy_p.r += 2 * dot(Eoe2*Tp * H1r, abs(vec2(t2op, t2ep)));
		energy_p.g += 2 * dot(Eoe2*Tp * H1g, abs(vec2(t2op, t2ep)));
		energy_p.b += 2 * dot(Eoe2*Tp * H1b, abs(vec2(t2op, t2ep)));
	}
	else {
		// Numeric

		const float start = 0.38f;
		const float end = 0.70f;
		const float lambda_step = (end-start)/float(m);

		energy_s = vec3(0);
		energy_p = vec3(0);
		vec3 weight = vec3(.0f);

		for (int s=0;s<m;++s) {
			// For 3 samples used predefined sample locations, otherwise sample at fixed intervals.
			const float lambdas3[3] = { 0.45, 0.55, 0.6 };
			const float lambda = m == 3 ? lambdas3[s] : (start + (float(s) + .5f) * lambda_step);

			const vec2 DPoo = numericSI(lambda, sin_theta, 1, 0, 1, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const vec2 DPoe = numericSI(lambda, sin_theta, 0, 1, 1, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const vec2 DPeo = numericSI(lambda, sin_theta, 1, 0, 0, 1, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const vec2 DPee = numericSI(lambda, sin_theta, 0, 1, 0, 1, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const complex_mat2 D = cmat2(r2oo*roo*DPoo + r2oe*reo*DPeo, r2oo*roe*DPoe + r2oe*ree*DPee,
								 		 r2eo*roo*DPoo + r2ee*reo*DPeo, r2eo*roe*DPoe + r2ee*ree*DPee);
			const complex_mat2 invD = cinv(cmat2_substract(cidentity(), D));

			const vec2 DPo = numericSI(lambda, sin_theta, 1, 0, 0, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const vec2 DPe = numericSI(lambda, sin_theta, 0, 1, 0, 0, OPLo, OPLe, OPLo2, OPLe2, Ioz, Iez, Ioz2, Iez2);
			const complex_mat2 D02 = cmat2(DPo, 	vec2(0),
										   vec2(0), DPe);

			const complex_mat2 I = cmat2_product(D02, invD);
			const vec2 energy = mag2(cvec2_by_mat2(vec2_by_cmat2(abs(vec2(t1o, t1e)), I), mat2(abs(t2os), abs(t2es), abs(t2op), abs(t2ep))));

			const float LX = X_1931(lambda*1000.f);
			const float LY = Y_1931(lambda*1000.f);
			const float LZ = Z_1931(lambda*1000.f);

			energy_s.x += LX * energy.x;
			energy_s.y += LY * energy.x;
			energy_s.z += LZ * energy.x;
			energy_p.x += LX * energy.y;
			energy_p.y += LY * energy.y;
			energy_p.z += LZ * energy.y;

			weight += vec3(LX, LY, LZ);
		}

		energy_s = XYZtoRGB(energy_s / weight);
		energy_p = XYZtoRGB(energy_p / weight);
	}



	// Reflection
	const float sR = mix(sqr(rss), sqr(rsp), ubo.polarization);			// Reflectivity to eye, s-polarized energy
	const float pR = mix(sqr(rps), sqr(rpp), ubo.polarization);			// Reflectivity to eye, p-polarized energy
	const vec3 R = texture(samplerEnv[ubo.envprobe],  i).rgb;
	// Refraction
	const vec3 T = texture(samplerEnv[ubo.envprobe], -o).rgb;
	energy_s = max(energy_s, vec3(0)) * T + sR * R;
	energy_p = max(energy_p, vec3(0)) * T + pR * R;

	const vec2 S = normalize((mat3(ubo.modelView) * X).xy);
	const vec2 P = normalize((mat3(ubo.modelView) * Y).xy);
	const vec3 energy_screen_s = energy_s;//S.x * energy_s + S.y * energy_p;
	const vec3 energy_screen_p = energy_p;//P.x * energy_s + P.y * energy_p;

	vec3 energy;
	// Polarization filters
	if (view_mode == 0)
		energy = energy_screen_s + energy_screen_p;
	else if (view_mode == 1)
		energy = energy_screen_s + 0.001f * energy_screen_p;
	else if (view_mode == 2)
		energy = 0.001f * energy_screen_s + energy_screen_p;

	outFragColor = vec4(energy, 1);
}
